--[[
Import libraries, other files.
Load data.
Main window.
Plugin's command.

DToX, 2015
Licensed under the same terms as Lua itself.
--]]





import "Turbine";
-- import "Turbine.UI";
import "Turbine.UI.Lotro";
import "Turbine.Gameplay"; -- ItemInfo





-- Variables and constants
--
local pCommandName = "mi";
local pName = "Mindon";
local totalItems;
local bActive = true;
local Results = {
   ["Found"] = {};
   ["Page"] = 1;
};
--[[ Data structure
   ["< Item's ID >"] = "< Name >";
   ...
--]]
local data_ENG;
local data_RUS;
local data;

local fonts = Turbine.UI.Lotro.Font;
local align = {}; -- <varName>.TopLeft or <varName>[1]
for k, v in pairs( Turbine.UI.ContentAlignment ) do
   align[k] = v;
   align[v] = v;
end

local LCRL = { "а","б","в","г","д","е","ё","ж","з","и","й","к","л","м","н","о","п","р","с","т","у","ф","х","ц","ч","ш","щ","ъ","ы","ь","э","ю","я" };
local UCRL = { "А","Б","В","Г","Д","Е","Ё","Ж","З","И","Й","К","Л","М","Н","О","П","Р","С","Т","У","Ф","Х","Ц","Ч","Ш","Щ","Ъ","Ы","Ь","Э","Ю","Я" };

local MAIN_WINDOW_WIDTH = 560;
local MAIN_WINDOW_HEIGHT = 540;
local MAIN_WINDOW_X = 300;
local MAIN_WINDOW_Y = 100;
-- Results list
local LIST_ITEM_WIDTH = 250;
local LIST_ITEM_HEIGHT = 36;
local LIST_ITEMS_NUMBER = 10;
local LIST_COLUMNS_NUMBER = 2;
local LIST_SCROLL_HEIGHT = 11;
local LIST_WIDTH = LIST_ITEM_WIDTH * LIST_COLUMNS_NUMBER;
local LIST_HEIGHT = LIST_ITEM_HEIGHT * LIST_ITEMS_NUMBER + LIST_SCROLL_HEIGHT;
local RESULTS_PER_PAGE = LIST_ITEMS_NUMBER * LIST_COLUMNS_NUMBER;

-- Forward declarations
--
local txt_search;
local lst_searchResults;
local lbl_searchStatus;
local btn_firstPage;
local btn_prevPage;
local btn_nextPage;
local btn_lastPage;
local txt_curPage;
local lbl_totalPages;





-- Functions
--
-- Shows messages in chat
local function wl( msg )
   Turbine.Shell.WriteLine( "<rgb=#FFFFFF>" .. pName .. ": " .. msg .. "</rgb>" );
end

-- Returns number of keys in table
local function tLen( t )
   if ( not t ) then
      return false;
   end

   local i = 0;
   for _ in pairs( t ) do
      i = i + 1;
   end
   return i;
end


-- Loads specified database
local function loadDatabase( name )
   local ok = pcall( import, "DToX.MindonL.Bases." .. name );
   if ( ok ) then
      local DB = DToX.MindonL.Bases[name];
      wl( ">> Загружена база " .. name .. ". Предметов: " .. tLen( DB ) .. "." );
      return DB;
   end
end

-- Converts russian letters to upper ones, otherwise uses standard function
local function toLower( str )
   local isRUS = false;
   for i, letter in pairs ( UCRL ) do
      if ( str:find( letter ) ) then
         str = str:gsub( letter, LCRL[i] );
         isRUS = true;
      end
   end
   -- Sanity check
   for i, letter in pairs ( LCRL ) do
      if ( str:find( letter ) ) then
         isRUS = true;
         break;
      end
   end

   if ( not isRUS ) then
      str = string.lower( str );
   end

   return str;
end

-- Converts to upper case and replaces some symbols for more accurate results
local RS = "[-+?'`]";
local function prepString( str )
   local str = toLower( str );
   if ( str:find( RS ) ) then
      str = str:gsub( RS, "_" );
   end
   return str;
end

-- Returns first word before space or original string
local function getWord( str )
   local p = str:find( " " );
   if ( p ) then
      local w = str:sub( 0, p - 1 );
      local s = str:sub( p + 1 );
      return w, s;
   else
      return str;
   end
end

-- Returns array of words. Delimiter: space
local function getWords( str )
   local str = str;
   local words = {};
   local word;
   local i = 0;
   while ( str ) do
      word, str = getWord( str );
      if ( word and word ~= "" ) then
         i = i + 1;
         words[i] = word;
      end
if ( i > 1000 ) then return; end
   end
   
   return words;
end

-- Returns coordinates of the top, left, bottom and right
-- borders of the object as an array with named keys
local function getBorders( obj )
   local width, height = obj:GetSize();
   local left, top = obj:GetPosition();
   local b = {
      ["Top"] = top;
      ["Left"] = left;
      ["Bottom"] = top + height;
      ["Right"] = left + width;
   };
   return b;
end

-- Creates and returns object with specified and predifined properties
local function newObject( parent, mode, width, text, fontSize )
   local fontSize = fontSize or 14;
   local fontMap = {
      [12] = fonts.Verdana12;
      [14] = fonts.Verdana14;
   };
   local font = fontMap[fontSize];

   if ( mode == "Button" ) then
      local button = Turbine.UI.Lotro.Button();
      button:SetParent( parent );
      button:SetFont( font );
      button:SetText( text );
      button:SetSize( width, 20 );
      return button;

   elseif ( mode == "Label" ) then
      local label = Turbine.UI.Label();
      label:SetParent( parent );
      label:SetFont( font );
      label:SetTextAlignment( align[4] );
      label:SetText( text );
      label:SetSize( width, 20 );
      return label;

   elseif ( mode == "Text" ) then
      local text = Turbine.UI.Lotro.TextBox();
      text:SetParent( parent );
      text:SetFont( font );
      text:SetSize( width, 20 );
      text:SetMultiline( false );
      return text;

   end
end



Results.startSearh = function()
   local q = txt_search:GetText();

   -- Try databases, until something is found
   for _, DB in pairs( data ) do
      lbl_searchStatus:SetText( "Раз ромашка, два ромашка..." );
      Results.search( q, DB );
      local n = tLen( Results.Found );
      if ( n > 0 ) then
         break;
      end
   end

   Results.changePage( 1 );
end

-- Searches items by name or ID.
-- Removes non-letter symbols from query. Converts query to words 
Results.search = function( str, DB )
   local items = DB;
   Results.Found = {};

   -- Try find by ID
   if ( items[str] ) then
      Results.Found[1] = str;
      return;
   end

   -- Get words from inputed string
   local cStr = prepString( str );
   local words = getWords( cStr );

   -- If name contains all words, add it to result
   for ID, name in pairs( items ) do
      local b = true;
      local n = prepString( name );
      for _, q in pairs( words ) do
         if ( not n:find( q ) ) then
            b = false;
            break;
         end
      end
      if ( b ) then
         local j = #Results.Found + 1;
         Results.Found[j] = ID;
      end
   end

end

-- Shows next, previous or specified page's results. Shows number of found items
Results.changePage = function( mode )
   local resultsNum = #Results.Found;
   local maxPages, remainder = math.modf( resultsNum / RESULTS_PER_PAGE );
   if ( remainder > 0 or resultsNum == 0 ) then
      maxPages = maxPages + 1;
   end

   -- Check mode's value and determine current page's number
   if ( mode == "+" ) then
      Results.Page = Results.Page + 1;
   elseif ( mode == "-" ) then
      Results.Page = Results.Page - 1;
   elseif ( mode == "Last" ) then
      Results.Page = maxPages;
   elseif ( mode ) then
      Results.Page = tonumber( mode );
   end
   -- Correct current page's number
   if ( Results.Page > maxPages ) then
      Results.Page = maxPages;
   elseif ( Results.Page < 1 ) then
      Results.Page = 1;
   end
   -- Update fields with page numbers
   txt_curPage:SetText( Results.Page );
   lbl_totalPages:SetText( maxPages );

   -- Toggle next and previous page buttons
   btn_firstPage:SetEnabled( true );
   btn_prevPage:SetEnabled( true );
   btn_nextPage:SetEnabled( true );
   btn_lastPage:SetEnabled( true );
   if ( Results.Page == 1 ) then
      btn_firstPage:SetEnabled( false );
      btn_prevPage:SetEnabled( false );
      if ( maxPages == 1 ) then
         btn_nextPage:SetEnabled( false );
         btn_lastPage:SetEnabled( false );
      end
   elseif ( Results.Page == maxPages ) then
      btn_nextPage:SetEnabled( false );
      btn_lastPage:SetEnabled( false );
   end

   lst_searchResults:ClearItems();
   local items = data_ENG;
   local foundItems = Results.Found;
   local resultStartNum = ( Results.Page - 1 ) * RESULTS_PER_PAGE;

   for i = 1, RESULTS_PER_PAGE do
      local j = resultStartNum + i;
      if ( j > resultsNum ) then
         break;
      end

      local ID = foundItems[j];
      local name = items[ID];

      -- Holds one list's item
      local carrier = Turbine.UI.Control();

      -- Shortcut
      local hex = string.format( "%X", ID );
      local sc = Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Item, "0x0,0x" .. hex );
      local itemInfo = sc:GetItem():GetItemInfo();

      -- Quickslot
      local itemQuickslot = Turbine.UI.Lotro.ItemInfoControl();
      itemQuickslot:SetParent( carrier );
      itemQuickslot:SetItemInfo( itemInfo );
      itemQuickslot:SetQuantity( 1 );
      -- itemQuickslot:SetSize( 36, 36 );

      -- Item's name (ENG)
      local itemName = Turbine.UI.Label();
      itemName:SetParent( carrier );
      itemName:SetSelectable( true );
      itemName:SetMultiline( false );
      itemName:SetFont( fonts.Verdana14 );
      itemName:SetText( name );
      local b = getBorders( itemQuickslot );
      local nameWidth = LIST_ITEM_WIDTH - b.Right - 10;
      itemName:SetSize( nameWidth, 15 );
      itemName:SetPosition( b.Right + 5, b.Top + 5 );

      -- Item's name (client's version)
      local itemNameTrans = Turbine.UI.Label();
      itemNameTrans:SetParent( carrier );
      itemNameTrans:SetSelectable( true );
      itemNameTrans:SetMultiline( false );
      itemNameTrans:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
      itemNameTrans:SetFont( fonts.Verdana14 );
      itemNameTrans:SetSize( nameWidth, 15 );
      local transName = data_RUS[ID];
      if ( transName ) then
         itemNameTrans:SetText( transName );
      else
         itemNameTrans:SetText( "<Отсутствует перевод>" );
      end
      local b = getBorders( itemName );
      itemNameTrans:SetPosition( b.Left, b.Bottom + 1 );

      carrier:SetSize( LIST_ITEM_WIDTH, LIST_ITEM_HEIGHT );
      function itemName:MouseEnter()
         if ( bActive ) then
            self:Focus();
            self:SelectAll();
         end
      end
      function itemName:MouseLeave()
         if ( bActive ) then
            self:Focus();
            self:DeselectAll();
         end
      end
      itemNameTrans.MouseEnter = itemName.MouseEnter;
      itemNameTrans.MouseLeave = itemName.MouseLeave;

      lst_searchResults:AddItem( carrier );
   end

   -- Show status
   local startNum = resultStartNum + 1;
   local endNum = resultStartNum + RESULTS_PER_PAGE;
   if ( endNum > resultsNum ) then
      endNum = resultsNum;
   end
   local str = "Найдено: " .. resultsNum .. " из " .. totalItems .. ".";
   if ( resultsNum > 0 ) then
      str = str .. " Показываются: с " .. startNum .. " по " .. endNum .. ".";
   end
   lbl_searchStatus:SetText( str );

end





-- Visuals
--
-- Main window
win_main = Turbine.UI.Lotro.Window();
win_main:SetSize( MAIN_WINDOW_WIDTH, MAIN_WINDOW_HEIGHT );
win_main:SetPosition( MAIN_WINDOW_X, MAIN_WINDOW_Y );
win_main:SetText( "Madness of... Mindon-Lite" );
win_main:SetVisible( true );
function win_main:Closed()
   wl( ">> Чтобы показать окно, используйте команду /mi или /mi show." );
end
function win_main:Activated()
   bActive = true;
end
function win_main:Deactivated()
   bActive = false;
end

-- Search. Info label
local lbl_search = newObject( win_main, "Label", 50, "Поиск:" );
lbl_search:SetTextAlignment( align[5] );
lbl_search:SetPosition( 30, 40 );

-- Search. Query field
txt_search = newObject( win_main, "Text", 200, "", 14 );
local b = getBorders( lbl_search );
txt_search:SetPosition( b.Right + 10, b.Top );
-- Start search when user presses Enter
function txt_search:KeyDown( args )
   local a = args.Action;
   if ( a == 162 ) then
   -- if ( args and args.Action and args.Action == 162 ) then
      Results.startSearh();
   end
end

-- Search. Results container
lst_searchResults = Turbine.UI.ListBox();
lst_searchResults:SetParent( win_main );
lst_searchResults:SetOrientation( Turbine.UI.Orientation.Vertical );
lst_searchResults:SetMaxItemsPerLine( LIST_ITEMS_NUMBER );
lst_searchResults:SetSize( LIST_WIDTH, LIST_HEIGHT );
lst_searchResults:SetBackColor( Turbine.UI.Color( 0.05, 0.05, 0.05 ) ); -- !!
local b = getBorders( lbl_search );
lst_searchResults:SetPosition( b.Left, b.Bottom + 10 );
local hScroll = Turbine.UI.Lotro.ScrollBar();
hScroll:SetParent( lst_searchResults );
hScroll:SetZOrder( 10 );
hScroll:SetOrientation( Turbine.UI.Orientation.Horizontal );
hScroll:SetSize( LIST_WIDTH, LIST_SCROLL_HEIGHT );
hScroll:SetPosition( 0, LIST_HEIGHT - LIST_SCROLL_HEIGHT );
lst_searchResults:SetHorizontalScrollBar( hScroll );

-- Search. Start button
local btn_startSearch = newObject( win_main, "Button", 60, "Найти" );
local b = getBorders( txt_search );
btn_startSearch:SetPosition( b.Right + 10, b.Top );
function btn_startSearch:MouseClick()
   Results.startSearh();
end

-- Search. Clear button
local btn_clearSearch = newObject( win_main, "Button", 80, "Очистить" );
local b = getBorders( btn_startSearch );
btn_clearSearch:SetPosition( b.Right + 10, b.Top );
function btn_clearSearch:MouseClick()
   txt_search:SetText( "" );
   lst_searchResults:ClearItems();
   lbl_searchStatus:SetText( "" );
   Results.search( "" );
   Results.changePage();
end

-- Search. Status label
lbl_searchStatus = newObject( win_main, "Label", 450, "" );
local b = getBorders( lst_searchResults );
lbl_searchStatus:SetPosition( b.Left, b.Bottom + 5 );


-- Search. First page button
btn_firstPage = newObject( win_main, "Button", 40, "<<" );
local b = getBorders( lbl_searchStatus );
btn_firstPage:SetPosition( b.Left, b.Bottom + 5 );
function btn_firstPage:MouseClick()
   Results.changePage( 1 );
end

-- Search. Previous page button
btn_prevPage = newObject( win_main, "Button", 40, "<" );
local b = getBorders( btn_firstPage );
btn_prevPage:SetPosition( b.Right + 5, b.Top );
function btn_prevPage:MouseClick()
   Results.changePage( "-" );
end

-- Search. Current page field
txt_curPage = newObject( win_main, "Text", 50, "1" );
txt_curPage:SetTextAlignment( align[5] );
local b = getBorders( btn_prevPage );
txt_curPage:SetPosition( b.Right + 10, b.Top );
function txt_curPage:TextChanged()
   local text = self:GetText();
   Results.changePage( text );
end

-- Search. Total pages label 
lbl_totalPages = newObject( win_main, "Label", 50, "" );
lbl_totalPages:SetTextAlignment( align[5] );
local b = getBorders( txt_curPage );
lbl_totalPages:SetPosition( b.Left, b.Bottom + 5 );

-- Search. Next page button 
btn_nextPage = newObject( win_main, "Button", 40, ">" );
local b = getBorders( txt_curPage );
btn_nextPage:SetPosition( b.Right + 10, b.Top );
function btn_nextPage:MouseClick()
   Results.changePage( "+" );
end

-- Search. Last page button 
btn_lastPage = newObject( win_main, "Button", 40, ">>" );
local b = getBorders( btn_nextPage );
btn_lastPage:SetPosition( b.Right + 5, b.Top );
function btn_lastPage:MouseClick()
   Results.changePage( "Last" );
end





-- Add command
--
pCommand = Turbine.ShellCommand();
Turbine.Shell.AddCommand( pCommandName, pCommand );

function pCommand:Execute( cmd, args )
   local cmd = "/" .. pCommandName;

   if ( args == "show" or args == "" ) then
      win_main:SetVisible( true );
      win_main:Activate();

   else
      wl( ">> Ошибка. Доступные команды:\n"
      .. cmd .. " или " .. cmd .. " show - показать окно плагина.\n"
      );
   end
end





-- Logic
--
-- Load databases
data_ENG = loadDatabase( "MindonDB_ENG" );
data_RUS = loadDatabase( "MindonDB_RUS" );
data = { data_ENG, data_RUS };
totalItems = tLen( data_ENG );

-- Just to remove scrollbar from empty list
lst_searchResults:AddItem( Turbine.UI.Label() );
